# ESU 2015, Leipzig
# Workshop: Comparing Corpora
# Peter Fankhauser
# compute distance measures and perform t-test on pairs of corpora represented
# as (smoothed) unigram language models

library(tm) # for constructing (sparse) document term matrices
library(slam) # for computing with sparse matrices
library(BSDA) # for t-test based on sample means and standard deviations

# read documents (one line per document)
# alternative: read document term matrices from CQP tabulate

docs <- read.table(file.choose(), header=T, fill=T, sep="\t", row.names=NULL, quote="")

#################### Preprocessing for one line per document #########################

# convert documents to a data frame
docs.df <- DataframeSource(as.matrix(docs$text))

# metadata (first 3 columns)
docs.meta <- docs[,1:3]

# construct the corpus from the data frame ds
docs.corpus <- Corpus(docs.df)

# construct a sparse document term matrix (dtm)
docs.dtm <- DocumentTermMatrix(docs.corpus)

#################### Alternative: Preprocessing for CQP tabulate output ################

# auxiliary functions for extracting corpus identifier (dir) and register (reg) from
# ids of the form "fr_F21_1"

splitid <- function(x) {
  m <- unlist(strsplit(x,"_"))
  dir <- m[1]
  reg <- substr(m[2],1,1)
  c(dir,reg)
}

docs.meta <- data.frame(t(mapply(as.character(docs[,1]),FUN=splitid)))
colnames(docs.meta) <- c("dir","reg")

docs.dtm <- as.simple_triplet_matrix(docs[,3:ncol(docs)])

####### end of preprocessing #######


# auxiliary vector for grouping by subcorpus (dir) and register (reg)
groupindex <- as.factor(paste(docs.meta$dir,docs.meta$reg,sep="_"))

# version for just grouping by subcorpus
# groupindex <- as.factor(docs$dir)

# sum up (absolute) frequencies for each subcorpus/register combination
docs.group.freq <-rollup(docs.dtm,1L,groupindex,na.rm=TRUE,fun=sum)

# sum up all frequencies  (for smoothing)
docs.all.freq <-rollup(docs.dtm,1L,fun=sum)

# calculate smoothed term probabilities
# simple smoothing with 0.005 of background corpus
lambda<-0.995

docs.group.prob <-as.matrix(docs.group.freq/row_sums(docs.group.freq))
docs.all.prob <-as.matrix(docs.all.freq/row_sums(docs.all.freq))

docs.group.smoothprob<-sweep(lambda*docs.group.prob,2,as.vector((1-lambda)*docs.all.prob),"+")

# some measures of divergence

# Kullback-Leibler Divergence

kld <- function(x,y) {
  if (y==0 && x > 0) {
    NaN
  }
  else if (x==0) {
    0
  }
  else {
  x*log2(x/y)
  }
}

# Jensen-Shannon Divergence

js <- function(x,y) {
  m <- (x+y)/2
  0.5*(kld(x,m)+kld(y,m))
}

# Jensen-Shannon Divergence ... individual terms

js.left <- function(x,y) {
  m <- (x+y)/2
  kld(x,m)
}

js.right <- function(x,y) {
  m <- (x+y)/2
  kld(y,m)
}

# Chisquare

chisquare <- function(o11,o12,o21,o22) {
  sum <- o11+o12+o21+o22
  e11 <- (o11+o12)*(o11+o21)/sum
  e12 <- (o11+o12)*(o12+o22)/sum
  e21 <- (o21+o22)*(o11+o21)/sum
  e22 <- (o21+o22)*(o12+o22)/sum
  (o11-e11)^2/e11+(o12-e12)^2/e12+(o21-e21)^2/e21+(o22-e22)^2/e22
}

# apply all divergence measures to a pair of corpora selected by sel1 and sel2

compute_divergences <- function (sel1,sel2) {
  prob1<-docs.group.smoothprob[sel1,,drop=FALSE]
  prob2<-docs.group.smoothprob[sel2,,drop=FALSE]
  kld1 <- mapply(kld,prob1,prob2)
  kld2 <- mapply(kld,prob2,prob1)
  js <- mapply(js,prob1,prob2)
  js.left <- mapply(js.left,prob1,prob2)
  js.right <- mapply(js.right,prob1,prob2)
  freq1 <- as.matrix(docs.group.freq[sel1,])
  freq2 <- as.matrix(docs.group.freq[sel2,])
  sum1 <- sum(freq1)
  sum2 <- sum(freq2)
  nfreq1 <- sum(freq1)-freq1
  nfreq2 <- sum(freq2)-freq2
  chisquare <- mapply(chisquare,freq1,freq2,nfreq1,nfreq2)
  pchi <- 1.0-pchisq(chisquare,1)
  res <- rbind(prob1,prob2,kld1,kld2,js,js.left,js.right,freq1,freq2,chisquare,pchi)
  rownames(res) <- c("p1","p2","kld1","kld2","js","js1","js2","f1","f2","chi","pchi")
  colnames(res) <- colnames(prob1)
  data.frame(t(res))
}

# selectors for subcorpora

sel1 <- "bl_A"
sel2 <- "bl_J"

divergences <- compute_divergences(sel1,sel2)
sdivergences <- divergences[order(-divergences$kld2),]

###### Significance Testing ######

# T-test

# sparse variants of mean and standard deviation
# x vector of values, nx number of 0 entries

sparsemean <- function(x,nx) {
  n <- nx+length(x)
  if (n<1) {
    NA
  }
  else {
    sum(x)/n
  }
}

sparsesd <- function(x,nx) {
  n <- nx+length(x)
  if (n<2) {
    NA
  }
  else {
    sum <- sum(x)
    sqrsum <- sum(x^2)
    sqrt((sqrsum-sum^2/n)/(n-1))
  }
}

# normalize frequencies by document length (unsmoothed probabilities)
docs.dtm.norm <- docs.dtm/row_sums(docs.dtm)
# means
docs.group.prob.mean<-as.matrix(rollup(docs.dtm.norm,1L,groupindex,EXPAND="sparse",FUN=sparsemean))
# standard deviations
docs.group.prob.sd<-as.matrix(rollup(docs.dtm.norm,1L,groupindex,EXPAND="sparse",FUN=sparsesd))
# number of documents per group
docs.group.count<-table(groupindex)

tsum.test.wrap <- function (m1,m2,s1,s2,n1,n2) {
  ttest<-tsum.test(m1, s.x=s1, n.x=n1, m2, s.y = s2,
                   n.y = n2, alternative = "two.sided", mu = 0, var.equal = FALSE,
                   conf.level = 0.95)
  ttest$p.value
}

compute_pvalues <- function(sel1,sel2) {
  means1 <- docs.group.prob.mean[sel1,,drop="FALSE"]
  means2 <- docs.group.prob.mean[sel2,,drop="FALSE"]
  sds1 <- docs.group.prob.sd[sel1,,drop="FALSE"]
  sds2 <- docs.group.prob.sd[sel2,,drop="FALSE"]
  ns1 <- as.vector(docs.group.count[sel1])
  ns2 <- as.vector(docs.group.count[sel2])
  res<-t(mapply(tsum.test.wrap,means1,means2,sds1,sds2,ns1,ns2))
  colnames(res)<-colnames(means1)
  rownames(res)<-c("pvalue")
  t(res)
}

pvalues <- compute_pvalues(sel1,sel2)

# combine all and sort/filter by various aspects

all <- cbind(divergences,pvalues)
all <- all[!is.nan(all$pvalue),]
sall <- all[order(-all$kld2),]
# filter everything with pvalue > 0.05
sall <- sall[sall$pvalue<0.05,]
chisig <- sall[sall$pchi>0.05,]







