use strict;

use File::Path qw(make_path remove_tree);
use Statistics::Distributions qw(tprob);
use POSIX;
use utf8;
use Encode;

# Author: Peter Fankhauser
# Description: read vrt-files to produce json data structures for contrastive visualization

# Usage: perl contrastjson-vrt.pl kld ../corpora/ememc/ememc_periods_config.txt ../corpora/ememc/ememc_periods ../corpora/ememc/ememc_periods_kld_json

# Input

my $divergence =
  $ARGV[0]
  ;    # kld for Kullback-Leibler Divergence, jsd for Jensen-Shannon Divergence

my $config =
  $ARGV[1];    # configuration for labels, mouseovers etc. in the visualization
my $dir1 =
  $ARGV[2]
  ; # input directory, consisting of directories for each main group, and subdirectories for each subgroup
my $dir2 = $ARGV[3];    # output directory for the json datastructures

# Some corpus specific configuration variables (mostly read from $config)

my $termindex =
  0;    # index for selecting term from a vrt line; should usually be 0.
my %mover     = ();    # mouse overs, read from config
my %revmover  = ();    # reverse mouse overs
my %variables = ();    # currently just corpus title
my %mapsort   = ();    # choose a particular order for (sub)corpora

# switch between Kullback-Leibler or Jensen-Shannon Divergence

my $kldflag = 1;
if ( $divergence eq "jsd" ) {
	$kldflag = 0;
}

# Read configuration

# changed to read (and require) UTF-8 encoding also for config

open( IN, "<:encoding(UTF-8)" , $config ) or die("can't open $config\n");
my $flag = 0;

while ( defined( my $line = <IN> ) ) {
	chomp($line);
	if ( $line =~ /TOOLTIPS/ ) {
		$flag = 1;
	}
	elsif ( $line =~ /VARIABLES/ ) {
		$flag = 2;
	}
	elsif ( $flag == 1 ) {
		my @fields  = split( /\t/, $line );
		my @fields1 = split( /,/,  $fields[1] );
		$mover{ $fields[0] }     = $fields1[0];
		$revmover{ $fields1[0] } = $fields[0];
		$mapsort{ $fields[0] }   = $fields1[1];
	}
	elsif ( $flag == 2 ) {
		my @fields = split( /\t/, $line );
		$variables{ $fields[0] } = $fields[1];
	}

}
close(IN);

# Processing

# Some (more) global variables

my $nfiles         = 0;     # number of files
my $ntokens        = 0;     # number of tokens

my %alltermcounts = ();

# make sure that the output directory indeed exists

make_path($dir2);

# recurse over directories

my %termcounts;
my @tokencounts;
my %totaltermcounts;
my $totaltokencount = 0;
my %localtermcounts = ();
my $localtokencount = 0;
my %sumrelfreq      = ();
my %sumsquarerelfreq =
  ()
  ; # for variance of relative frequency of term $term in subcorpus $i (per document)
my @docnum;

my $lambda = 0.05
  ; # mixing parameter for Jelinek-Mercer smoothing subcorpus frequencies with corpus frequencies
my $maxlength = 25;
my $minoccur  = 10;
my $maxterms  = 250
  ; # maximum number of terms for a particular contrast, is further filtered by tprob (significance)
my @mindist = ( 100, 100 );
my @maxdist = ( 0,   0 );

my @siglevels = ( 0.01, 0.05, 0.10, 0.25, 0.50 );
my @sigcounts = ();

# record overall maximum/minimum term distances and frequencies
# for minimum term distances and frequencies record minimum of top $step, 2*$step <= $maxterms
my @tmaxdist  = ( 0, 0 );
my @tmindist  = ();
my @tmaxfreq  = ( 0, 0 );
my @tminfreq  = ();
my $step      = 25;
my $stepcount = 0;

for ( my $i = 0 ; $i < 2 ; $i++ ) {
	for ( my $j = 0 ; $j < $maxterms / $step ; $j++ ) {
		$tmindist[$i][$j] = 10;
		$tminfreq[$i][$j] = 1;
	}
}

# first, process all docs for an overall termcount

collectallterms($dir1);

print "tokens: $ntokens, docs: $nfiles\n";

# then, process top directories (for left heatmap)

print "Level 1\n";
my @index =
  processfiles( $dir1, $dir1, $dir2, 0, 1, $variables{"corpustitle"} );

# next process all pairs of subdirectories (for middle and right heatmap);

print "Level 2\n";
opendir( CURR, $dir1 ) or die("can't open $dir1");
my @subdirs = sortdirs( grep( /[A-Za-z0-9_]/, readdir(CURR) ) );
closedir(CURR);

for ( my $i = 0 ; $i < scalar(@subdirs) ; $i++ ) {
	for ( my $j = 0 ; $j < scalar(@subdirs) ; $j++ ) {
		my $sameflag = 0;
		if ( $i == $j ) {
			$sameflag = 1;
		}
		my $subdir1 = $dir1 . "/" . $subdirs[$i];
		my $subdir2 = $dir1 . "/" . $subdirs[$j];
		my $outdir  = $dir2 . "/" . sprintf( "dir_%03d_%03d", $i, $j );
		print "$i $subdir1 $j $subdir2 $outdir\n";
		my $title = $subdirs[$i];
		if ( $i != $j ) {
			$title .= " vs. " . $subdirs[$j];
		}
		processfiles( $subdir1, $subdir2, $outdir, 1, $sameflag, $title );
	}
}

# write min and max (for calibrating colors etc. of level 1 comparison - left heatmap - and the two level 2 comparisons)

my $out = $dir2 . "/minmax.json";
open( OUT, ">" . $out ) or die("can't open $out");
binmode( OUT, ":utf8" );

print OUT "[\n";
for ( my $i = 0 ; $i < 2 ; $i++ ) {
	printf OUT "{\"cdist\":\n{\"min\":%5.4f,\n\"max\":%5.4f},\n", $mindist[$i],
	  $maxdist[$i];
	print OUT "\"tdist\":\n{\"min\":[";
	my $tstring = "";
	for ( my $j ; $j < $maxterms / $step ; $j++ ) {
		$tstring .= sprintf( "%9.8f,", $tmindist[$i][$j] );
	}
	$tstring =~ s/,$//;
	printf OUT "%s],\n\"max\":%9.8f},\n", $tstring, $tmaxdist[$i];
	print OUT "\"tfreq\":\n{\"min\":[";
	$tstring = "";
	for ( my $j ; $j < $maxterms / $step ; $j++ ) {
		$tstring .= sprintf( "%9.8f,", $tminfreq[$i][$j] );
	}
	$tstring =~ s/,$//;
	printf OUT "%s],\n\"max\":%9.8f}\n}", $tstring, $tmaxfreq[$i];

	if ( $i == 0 ) {
		print OUT ",\n";
	}
	else {
		print OUT "]\n";
	}
}
close(OUT);

# End of "main"

sub processfiles {
	my $dir11    = shift;
	my $dir12    = shift;
	my $dir2     = shift;
	my $dirflag  = shift;
	my $sameflag = shift;
	my $title    = shift;

	# we're helplessly working with global datastructures here...
	%termcounts       = ();
	@tokencounts      = ();
	%totaltermcounts  = ();
	$totaltokencount  = 0;
	%sumrelfreq       = ();
	%sumsquarerelfreq = ();
	@docnum           = ();

	# the directories are not necessarily identical
	opendir( CURR, $dir11 ) or die("can't open $dir11");
	my @subdirs1 = sortdirs( grep( /[A-Za-z0-9_]/, readdir(CURR) ) );
	closedir(CURR);
	my $index1 = 0;
	foreach my $subdir (@subdirs1) {
		collectterms( $dir11 . "/" . $subdir, $index1 );
		$index1++;
	}

	# note that with $sameflag = 1 and $dir11 = $dir12, the
	# totalcounts are added up twice.
	opendir( CURR, $dir12 ) or die("can't open $dir12");
	my @subdirs2 = sortdirs( grep( /[A-Za-z0-9_]/, readdir(CURR) ) );
	closedir(CURR);
	my $index2 = 0;
	foreach my $subdir (@subdirs2) {
		collectterms( $dir12 . "/" . $subdir, $index1 + $index2 );
		$index2++;
	}

	# we have termcounts and tokencounts
	# produce distance matrix, significance matrices, and termweights

	my @matrix  = ();
	my @tmatrix = ();

	# for counting the number of terms with tprob < significance level
	# 5 matrices for significance levels 0.01, 0.05, 0.10, 0.25, 0.50
	my @sigmatrix = ();

	my %termweights;
	my %tvalues;     # t-values for a Welch-ttest (unpaired, unequal variances)
	my %dfvalues;    # degrees of freedom.

	for ( my $i = 0 ; $i < $index1 ; $i++ ) {
		# some of the calculations in the $j loop depending only on $i could be
		# be put here.
		for ( my $j = 0 ; $j < $index2 ; $j++ ) {
			if (
				   $tokencounts[$i] > 0
				&& $tokencounts[ $index1 + $j ] > 0
				&& (   $sameflag == 0
					|| $i != $j
					|| $totaltokencount / 2.0 - $tokencounts[$i] > 0 )
			  )
			{
				foreach my $term ( keys(%termcounts) ) {
					my $p  = $totaltermcounts{$term} / $totaltokencount;
					my $q1 =
					  ( 1 - $lambda ) * $termcounts{$term}[$i] /
					  $tokencounts[$i] + $lambda * $p;
					my $q2 =
					  ( 1 - $lambda ) * $termcounts{$term}[ $index1 + $j ] /
					  $tokencounts[ $index1 + $j ] + $lambda * $p;
					my $diff = 0;
					if ($kldflag) {
						$diff = $q1 * ( log2($q1) - log2($q2) );
					}
					else {
						$diff = js( $q1, $q2 );
					}
					$matrix[$i][$j] += $diff;
					if ( $i == $j && $sameflag == 1 ) {

# for the main diagonal, we record the termweight of $i vs. the rest of the corpus.
# total counts need to be divided by 2, because they were added up twice.
						my $othertermcount = $totaltermcounts{$term} / 2.0 -
						  $termcounts{$term}[$i];
						my $othertokencount =
						  $totaltokencount / 2.0 - $tokencounts[$i];
						my $q21 =
						  ( 1 - $lambda ) * $othertermcount / $othertokencount +
						  $lambda * $p;
						my $diff1 = 0;
						if ($kldflag) {
							$diff1 = $q1 * ( log2($q1) - log2($q21) );
						}
						else {
							$diff1 = jsasym( $q1, $q21 );
						}
						$termweights{$term}[$i][$i] = $diff1;

						my $sum2       = 0;
						my $sumsquare2 = 0;
						my $n2         = 0;
						for ( my $k = 0 ; $k < $index1 ; $k++ ) {
							if ( $k != $j ) {
								$sum2       += $sumrelfreq{$term}[$k];
								$sumsquare2 += $sumsquarerelfreq{$term}[$k];
								$n2         += $docnum[$k];
							}
						}
						my @ttest = ttest(
							$docnum[$i], $n2, $sumrelfreq{$term}[$i],
							$sum2, $sumsquarerelfreq{$term}[$i],
							$sumsquare2, $term
						);
						$tvalues{$term}[$i][$i]  = $ttest[0];
						$dfvalues{$term}[$i][$i] = $ttest[1];
						my $tprob = tprob0( floor( $dfvalues{$term}[$i][$i] ),
							$tvalues{$term}[$i][$i] );
					}
					else {

					   # otherwise the termweight is the kld factor of $i vs. $j
						if ($kldflag) {
							$termweights{$term}[$i][$j] = $diff;
						}
						else {
							$termweights{$term}[$i][$j] = jsasym( $q1, $q2 );
						}
						my @ttest = ttest(
							$docnum[$i],
							$docnum[ $index1 + $j ],
							$sumrelfreq{$term}[$i],
							$sumrelfreq{$term}[ $index1 + $j ],
							$sumsquarerelfreq{$term}[$i],
							$sumsquarerelfreq{$term}[ $index1 + $j ],
							$term
						);
						$tvalues{$term}[$i][$j]  = $ttest[0];
						$dfvalues{$term}[$i][$j] = $ttest[1];
						my $tprob = tprob0( floor( $dfvalues{$term}[$i][$j] ),
							$tvalues{$term}[$i][$j] );

					}
				}
				$mindist[$dirflag] = min( $mindist[$dirflag], $matrix[$i][$j] );
				$maxdist[$dirflag] = max( $maxdist[$dirflag], $matrix[$i][$j] );
			}
			else {
				$matrix[$i][$j] = -1;
			}
		}
	}

	# produce output

	# the termweights
	# and generate the subdirectories

	for ( my $i = 0 ; $i < $index1 ; $i++ ) {
		for ( my $j = 0 ; $j < $index2 ; $j++ ) {
			$tmatrix[$i][$j] = 1.0;

			# copy termweights for sorting
			my %termweights1 = ();
			foreach my $term ( keys(%termweights) ) {
				$termweights1{$term} = $termweights{$term}[$i][$j];
			}
			my @sortedterms =
			  sort { $termweights1{$b} <=> $termweights1{$a} }
			  keys %termweights1;
			open( OUT,
				    ">" . $dir2
				  . "/terms_"
				  . sprintf( "%03d_%03d", $i, $j )
				  . ".json" ) or die("can't open terms $i $j");
			binmode( OUT, ":utf8" );
			my $termtitle = $subdirs1[$i];
			my @pars1     = split( /\//, $dir11 );
			my $par1      = $pars1[-1];
			if ( $dirflag == 0 && $i == $j ) {
				$termtitle .= " vs. Rest";
			}
			elsif ( $dirflag == 0 ) {
				$termtitle .= " vs. " . $subdirs2[$j];
			}
			elsif ( $i == $j && $sameflag == 1 ) {
				$termtitle .= " vs. Rest (" . $par1 . ")";
			}
			else {
				my @pars2 = split( /\//, $dir12 );
				my $par2 = $pars2[-1];
				$termtitle .=
				  " (" . $par1 . ") vs. " . $subdirs2[$j] . " (" . $par2 . ")";
			}
			print OUT "{\n\"title\": \"$termtitle\",\n\"terms\": [\n";
			if (
				   $tokencounts[$i] > 0
				&& $tokencounts[ $index1 + $j ] > 0
				&& (   $sameflag == 0
					|| $i != $j
					|| $totaltokencount / 2.0 - $tokencounts[$i] > 0 )
			  )
			{
				my $stepcount = 0;
				@sigcounts = ();
				for ( my $k = 0 ; $k < $maxterms ; $k++ ) {
					my $weight = $termweights1{ $sortedterms[$k] };
					my $freq   =
					  $termcounts{ $sortedterms[$k] }[$i] / $tokencounts[$i];
					if ( ( $k + 1 ) % $step == 0 ) {
						$tmindist[$dirflag][$stepcount] =
						  min( $tmindist[$dirflag][$stepcount], $weight );
						$tminfreq[$dirflag][$stepcount] =
						  min( $tminfreq[$dirflag][$stepcount], $freq );
						$stepcount++;
					}
					$tmaxdist[$dirflag] = max( $tmaxdist[$dirflag], $weight );
					$tmaxfreq[$dirflag] = max( $tmaxfreq[$dirflag], $freq );
					my $tprob = tprob0(
						floor( $dfvalues{ $sortedterms[$k] }[$i][$j] ),
						$tvalues{ $sortedterms[$k] }[$i][$j]
					);
					addupsiglevels($tprob);
					printf OUT "[\"%s\", %9.8f, %9.8f, %9.8f]",
					  $sortedterms[$k], $weight, $freq, $tprob;
					$tmatrix[$i][$j] = min( $tmatrix[$i][$j], $tprob );
					if ( $k < $maxterms - 1 ) {
						print OUT ",\n";
					}
				}
				for ( my $k = 0 ; $k < scalar(@siglevels) ; $k++ ) {
					$sigmatrix[$k][$i][$j] = $sigcounts[$k];
				}
			}
			print OUT "\n]\n}";
			close OUT;
			if ( $dirflag == 0 ) {
				make_path( $dir2 . "/dir_" . sprintf( "%03d_%03d", $i, $j ) );
			}
		}

		# the matrix

		my $out = $dir2 . "/matrix.json";
		if ( !( -d $dir2 ) ) {
			make_path($dir2);
		}
		open( OUT, ">" . $out ) or die("can't open $out");
		binmode( OUT, ":utf8" );
		# try to make again sure that $title contains no newline - probably redundant.
		chomp($title);
		print OUT "{\n\"title\": \"$title\",\n";
		my $labels = "";
		my $movers = "";
		for ( my $i = 0 ; $i < $index1 ; $i++ ) {
			$labels .= '"' . renamelabel( $subdirs1[$i] ) . '",';
			$movers .= '"' . $mover{ renamelabel( $subdirs1[$i] ) } . '",';
		}
		$labels =~ s/,$//;
		$movers =~ s/,$//;
		print OUT "\"xlabels\": [$labels],\n\"xmover\": [$movers],\n";
		$labels = "";
		$movers = "";
		for ( my $i = 0 ; $i < $index2 ; $i++ ) {
			$labels .= '"' . renamelabel( $subdirs2[$i] ) . '",';
			$movers .= '"' . $mover{ renamelabel( $subdirs2[$i] ) } . '",';
		}
		$labels =~ s/,$//;
		$movers =~ s/,$//;
		print OUT "\"ylabels\": [$labels],\n\"ymover\": [$movers],\n";
		print OUT "\"matrix\": [\n";
		for ( my $i = 0 ; $i < $index1 ; $i++ ) {
			my $values = "";
			for ( my $j = 0 ; $j < $index2 ; $j++ ) {
				$values .= sprintf( "%5.4f,", $matrix[$i][$j] );
			}
			$values =~ s/,$//;
			print OUT "[$values]";
			if ( $i < $index1 - 1 ) {
				print OUT ",";
			}
			print OUT "\n";
		}
		print OUT "],\n\"tmatrix\": [\n";
		for ( my $i = 0 ; $i < $index1 ; $i++ ) {
			my $values = "";
			for ( my $j = 0 ; $j < $index2 ; $j++ ) {
				$values .= sprintf( "%9.8f,", $tmatrix[$i][$j] );
			}
			$values =~ s/,$//;
			print OUT "[$values]";
			if ( $i < $index1 - 1 ) {
				print OUT ",";
			}
			print OUT "\n";
		}
		for ( my $k = 0 ; $k < scalar(@siglevels) ; $k++ ) {
			print OUT "],\n\"sigmatrix$k\": [\n";
			for ( my $i = 0 ; $i < $index1 ; $i++ ) {
				my $values = "";
				for ( my $j = 0 ; $j < $index2 ; $j++ ) {
					$values .= sprintf( "%d,", $sigmatrix[$k][$i][$j] );
				}
				$values =~ s/,$//;
				print OUT "[$values]";
				if ( $i < $index1 - 1 ) {
					print OUT ",";
				}
				print OUT "\n";
			}
		}
		print OUT "]\n}\n";
		close OUT;
	}
	return ( $index1, $index2 );
}

sub collectallterms {
	my $dir = $_[0];
	if ( -d $dir ) {
		print "$nfiles $dir\n";
		opendir( CURR, $dir );
		my @subdirs = sortdirs( grep( /[A-Za-z0-9_]/, readdir(CURR) ) );
		closedir(CURR);
		foreach my $subdir (@subdirs) {
			if ( $subdir =~ /^[\s]*$/ ) { }
			else {
				collectallterms( $dir . "/" . $subdir );
			}
		}
	}
	else {

		# my $in = "<" . $dir;
		open( IN, "<:encoding(UTF-8)", $dir ) or die("can't open $dir");
		$nfiles++;
		while ( defined( my $line = <IN> ) ) {
			chomp($line);
			if ( $line =~ /^<.*?>/ ) { }
			else {
				my $term = extractfeature($line);
				if ( $term ne "" ) {
					$alltermcounts{$term}++;
					$ntokens++;
				}
			}
		}
		close(IN);
	}
}

sub collectterms {
	my $dir   = $_[0];
	my $index = $_[1];
	if ( -d $dir ) {
		print "$dir\n";
		opendir( CURR, $dir );
		my @subdirs = sortdirs( grep( /[A-Za-z0-9_]/, readdir(CURR) ) );
		closedir(CURR);
		foreach my $subdir (@subdirs) {
			if ( $subdir =~ /^[\s]*$/ ) { }
			else {
				collectterms( $dir . "/" . $subdir, $index );
			}
		}
	}
	else {

		# my $in = "<" . $dir;
		open( IN, "<:encoding(UTF-8)", $dir ) or die("can't open $dir");
		my $oldterm = "";

		# auxiliary structure for calculating local relative frequencies

		%localtermcounts = ();
		$localtokencount = 0;
		my %lines = ();
		while ( defined( my $line = <IN> ) ) {
			chomp($line);
			if ( $line =~ /^<.*?>/ ) { }
			else {
				my $term = extractfeature($line);
				if ( $term ne "" ) {
					addterm( $term, $index );
				}
			}
		}
		close(IN);
		if ( $localtokencount > 0 ) {
			foreach my $term ( keys %totaltermcounts ) {
				my $relfreq = $localtermcounts{$term} / $localtokencount;
				$sumrelfreq{$term}[$index]       += $relfreq;
				$sumsquarerelfreq{$term}[$index] += $relfreq**2;
			}
			$docnum[$index]++;
		}
	}
}

sub log2 {
	my $n = shift;
	return log($n) / log(2);
}

sub renamelabel {
	my $label = shift;

	#	return $revmover{$label};
	return $label;
}

sub min {
	my $x = shift;
	my $y = shift;
	if ( $x == 0 ) {
		return $y;
	}
	if ( $y == 0 ) {
		return $x;
	}
	if ( $x < $y ) {
		return $x;
	}
	else {
		return $y;
	}
}

sub max {
	my $x = shift;
	my $y = shift;
	if ( $x < $y ) {
		return $y;
	}
	else {
		return $x;
	}
}

sub addterm {
	my $term  = shift;
	my $index = shift;
	if ( $alltermcounts{$term} >= $minoccur ) {
		$tokencounts[$index]++;
		$localtokencount++;
		$termcounts{$term}[$index]++;
		$totaltokencount++;
		$totaltermcounts{$term}++;
		$localtermcounts{$term}++;
	}
}

sub ttest {
	my $n1         = shift;
	my $n2         = shift;
	my $sum1       = shift;
	my $sum2       = shift;
	my $sumsquare1 = shift;
	my $sumsquare2 = shift;
	my $term       = shift;
	if ( $n1 <= 1 || $n2 <= 1 ) {
		return ( 0, 0 );
	}
	my $mean1 = $sum1 / $n1;
	my $mean2 = $sum2 / $n2;
	my $var1  = ( $sumsquare1 - ( $sum1**2 / $n1 ) ) / ( $n1 - 1 );
	my $var2  = ( $sumsquare2 - ( $sum2**2 / $n2 ) ) / ( $n2 - 1 );
	if ( $mean1 == 0 && $mean2 == 0 ) {
		return ( 0, 0 );
	}
	my $x1 = $var1 / $n1;
	my $x2 = $var2 / $n2;
	if ( $x1 + $x2 == 0 ) {
		print "$term $mean1 $mean2\n";
		return ( 0, 0 );
	}
	my $tvalue = ( $mean1 - $mean2 ) / ( sqrt($x1) + sqrt($x2) );
	my $dfvalue =
	  ( ( $x1 + $x2 )**2 ) / ( $x1**2 / ( $n1 - 1 ) + $x2**2 / ( $n2 - 1 ) );
	if ( $dfvalue > $n1 + $n2 - 2 ) {
		print
		  "warning: weird $dfvalue $term $mean1 $mean2 $var1 $var2 $n1 $n2\n";
	}
	$dfvalue = min( $n1 + $n2 - 2, floor($dfvalue) );
	return ( $tvalue, $dfvalue );
}

sub js {
	my $p = shift;
	my $q = shift;
	my $m = ( $p + $q ) / 2;
	return ( $p * log2( $p / $m ) + $q * log2( $q / $m ) ) / 2;
}

# for term weighting with JS-Divergence, just return the divergence between $p and the
# mean distribution $m.

sub jsasym {
	my $p = shift;
	my $q = shift;
	my $m = ( $p + $q ) / 2;
	return $p * log2( $p / $m );
}

sub tprob0 {
	my $arg1 = shift;
	my $arg2 = shift;
	if ( $arg1 == 0 || $arg2 == 0 ) {
# kind of arbitrary, in principle tprob is undefined in such a case.
		return 0.5;
	}
	else {
		return tprob( $arg1, $arg2 );
	}
}

sub sortdirs {
	my @list = @_;
	foreach my $name (@list) {
		if ( not( exists $mapsort{$name} ) ) {
			return @list;
			last();
		}
	}
	my @ret = ();
	foreach my $name (@list) {
		$ret[ $mapsort{$name} ] = $name;
	}
	return @ret;
}

sub addupsiglevels {
	my $tprob = shift;
	for ( my $i = 0 ; $i < scalar(@siglevels) ; $i++ ) {
		if ( $tprob < $siglevels[$i] ) { $sigcounts[$i]++; }
	}
}

# extract and cleanup terms, needs to be handled with care to avoid no result queries
# from the word clouds.
# adapt this to (e.g.) exclude stop words, work on lemmas, select only particular POS etc.

sub extractfeature {
	my $line = shift;
	my @features = split( /\t/, $line );
	my $term = $features[$termindex];
	$term =~ s/[^\p{XPosixAlnum}.,\/~']//g;
	$term =~ s/\/IN$//;
	$term =~ s/[.,\/]//g;
	$term = lc($term);
}

